@@ -41,6 +41,7 @@
 #include "libdecor.h"
 #include "utils.h"
 #include "cursor-settings.h"
+#include "os-compatibility.h"
 
 #include "xdg-shell-client-protocol.h"
 
@@ -105,6 +106,8 @@ struct window {
 	int content_height;
 	int configured_width;
 	int configured_height;
+	int floating_width;
+	int floating_height;
 	enum libdecor_window_state window_state;
 	struct wl_list outputs;
 	int scale;
@@ -173,12 +176,15 @@ resize(struct window *window, int width, int height)
 	}
 
 	/* commit changes to decorations */
-	state = libdecor_state_new( width, height);
+	state = libdecor_state_new(width, height);
 	libdecor_frame_commit(window->frame, state, NULL);
 	libdecor_state_free(state);
 	/* force redraw of content and commit */
 	window->configured_width = width;
 	window->configured_height = height;
+	/* store floating dimensions */
+	window->floating_width = width;
+	window->floating_height = height;
 	redraw(window);
 }
 
@@ -650,8 +656,7 @@ keyboard_key(void *data,
 				printf("set fixed-size\n");
 				libdecor_frame_unset_capabilities(window->frame,
 							LIBDECOR_ACTION_RESIZE);
-			}
-			else {
+			} else {
 				printf("set resizeable\n");
 				libdecor_frame_set_capabilities(window->frame,
 							LIBDECOR_ACTION_RESIZE);
@@ -981,32 +986,6 @@ static const struct wl_buffer_listener buffer_listener = {
 	buffer_release
 };
 
-static int
-create_anonymous_file(off_t size)
-{
-	int fd;
-
-	int ret;
-
-	fd = memfd_create("libdecor-demo", MFD_CLOEXEC | MFD_ALLOW_SEALING);
-
-	if (fd < 0)
-		return -1;
-
-	fcntl(fd, F_ADD_SEALS, F_SEAL_SHRINK);
-
-	do {
-		ret = posix_fallocate(fd, 0, size);
-	} while (ret == EINTR);
-	if (ret != 0) {
-		close(fd);
-		errno = ret;
-		return -1;
-	}
-
-	return fd;
-}
-
 static struct buffer *
 create_shm_buffer(int width,
 		  int height,
@@ -1020,7 +999,7 @@ create_shm_buffer(int width,
 	stride = width * 4;
 	size = stride * height;
 
-	fd = create_anonymous_file(size);
+	fd = os_create_anonymous_file(size);
 	if (fd < 0) {
 		fprintf(stderr, "creating a buffer file for %d B failed: %s\n",
 			size, strerror(errno));
@@ -1121,8 +1100,8 @@ handle_configure(struct libdecor_frame *frame,
 		height = window->content_height;
 	}
 
-	width = (width == 0) ? DEFAULT_WIDTH : width;
-	height = (height == 0) ? DEFAULT_HEIGHT : height;
+	width = (width == 0) ? window->floating_width : width;
+	height = (height == 0) ? window->floating_height : height;
 
 	window->configured_width = width;
 	window->configured_height = height;
@@ -1137,6 +1116,12 @@ handle_configure(struct libdecor_frame *frame,
 	libdecor_frame_commit(frame, state, configuration);
 	libdecor_state_free(state);
 
+	/* store floating dimensions */
+	if (libdecor_frame_is_floating(window->frame)) {
+		window->floating_width = width;
+		window->floating_height = height;
+	}
+
 	redraw(window);
 }
 
@@ -1239,10 +1224,19 @@ int
 main(int argc,
      char **argv)
 {
+	struct timeval start;
+	struct timeval now;
 	struct wl_display *wl_display;
 	struct wl_registry *wl_registry;
 	struct libdecor *context;
 	struct output *output;
+	int timeout = -1;
+	bool timed_quit = false;
+
+	if (argc > 1 && strcmp(argv[1], "--timed-quit") == 0) {
+		timed_quit = true;
+		timeout = 500; /* ms */
+	}
 
 	/* write all output to stdout immediately */
 	setbuf(stdout, NULL);
@@ -1277,6 +1271,10 @@ main(int argc,
 	window->wl_surface = wl_compositor_create_surface(wl_compositor);
 	wl_surface_add_listener(window->wl_surface, &surface_listener, window);
 
+	/* initialise content dimensions */
+	window->floating_width = DEFAULT_WIDTH;
+	window->floating_height = DEFAULT_HEIGHT;
+
 	context = libdecor_new(wl_display, &libdecor_iface);
 	window->frame = libdecor_decorate(context, window->wl_surface,
 					  &libdecor_frame_iface, window);
@@ -1284,7 +1282,18 @@ main(int argc,
 	libdecor_frame_set_title(window->frame, "libdecor demo");
 	libdecor_frame_map(window->frame);
 
-	while (libdecor_dispatch(context, -1) >= 0);
+	gettimeofday(&start, NULL);
+	libdecor_frame_set_min_content_size(window->frame, 15 * chk, 10 * chk);
+
+	while (libdecor_dispatch(context, timeout) >= 0) {
+		if (timed_quit) {
+			gettimeofday(&now, NULL);
+			if (now.tv_sec >= start.tv_sec + 2) {
+				fprintf(stderr, "Exiting due to --timed-quit\n");
+				libdecor_frame_close(window->frame);
+			}
+		}
+	}
 
 	free_outputs();
 
