@@ -39,6 +39,7 @@
 #include "libdecor-plugin.h"
 #include "utils.h"
 #include "cursor-settings.h"
+#include "os-compatibility.h"
 
 #include <cairo/cairo.h>
 #include <pango/pangocairo.h>
@@ -110,7 +111,8 @@ streql(const char *str1, const char *str2)
 enum decoration_type {
 	DECORATION_TYPE_NONE,
 	DECORATION_TYPE_ALL,
-	DECORATION_TYPE_TITLE_ONLY
+	DECORATION_TYPE_MAXIMIZED,
+	DECORATION_TYPE_TILED
 };
 
 enum component {
@@ -537,32 +539,6 @@ libdecor_plugin_cairo_frame_new(struct libdecor_plugin *plugin)
 	return &frame_cairo->frame;
 }
 
-static int
-create_anonymous_file(off_t size)
-{
-	int ret;
-
-	int fd;
-
-	fd = memfd_create("libdecor-cairo", MFD_CLOEXEC | MFD_ALLOW_SEALING);
-
-	if (fd < 0)
-		return -1;
-
-	fcntl(fd, F_ADD_SEALS, F_SEAL_SHRINK);
-
-	do {
-		ret = posix_fallocate(fd, 0, size);
-	} while (ret == EINTR);
-	if (ret != 0) {
-		close(fd);
-		errno = ret;
-		return -1;
-	}
-
-	return fd;
-}
-
 static void
 toggle_maximized(struct libdecor_frame *const frame)
 {
@@ -610,7 +586,7 @@ create_shm_buffer(struct libdecor_plugin_cairo *plugin_cairo,
 	stride = buffer_width * 4;
 	size = stride * buffer_height;
 
-	fd = create_anonymous_file(size);
+	fd = os_create_anonymous_file(size);
 	if (fd < 0) {
 		fprintf(stderr, "creating a buffer file for %d B failed: %s\n",
 			size, strerror(errno));
@@ -1020,12 +996,16 @@ ensure_component(struct libdecor_frame_cairo *frame_cairo,
 static void
 ensure_border_surfaces(struct libdecor_frame_cairo *frame_cairo)
 {
+	int min_width, min_height;
+
 	frame_cairo->shadow.opaque = false;
 	ensure_component(frame_cairo, &frame_cairo->shadow);
 
+	libdecor_frame_get_min_content_size(&frame_cairo->frame,
+					    &min_width, &min_height);
 	libdecor_frame_set_min_content_size(&frame_cairo->frame,
-					    MAX(56, 4 * BUTTON_WIDTH),
-					    MAX(56, TITLE_HEIGHT + 1));
+		MAX(min_width, (int)MAX(56, 4 * BUTTON_WIDTH)),
+		MAX(min_height, (int)MAX(56, TITLE_HEIGHT + 1)));
 }
 
 static void
@@ -1134,10 +1114,16 @@ draw_title_text(struct libdecor_frame_cairo *frame_cairo,
 	double text_x, text_y;
 	double text_width, text_height;
 
+	const char *title;
+
+	title = libdecor_frame_get_title((struct libdecor_frame*) frame_cairo);
+	if (!title)
+		return;
+
 	layout = pango_cairo_create_layout(cr);
 
 	pango_layout_set_text(layout,
-			      libdecor_frame_get_title((struct libdecor_frame*) frame_cairo),
+			      title,
 			      -1);
 	pango_layout_set_font_description(layout, frame_cairo->plugin_cairo->font);
 	pango_layout_get_size(layout, &text_extents_width, &text_extents_height);
@@ -1249,14 +1235,15 @@ draw_component_content(struct libdecor_frame_cairo *frame_cairo,
 	case NONE:
 		break;
 	case SHADOW:
-		render_shadow(cr,
-			      frame_cairo->shadow_blur,
-			      -(int)SHADOW_MARGIN/2,
-			      -(int)SHADOW_MARGIN/2,
-			      width + SHADOW_MARGIN,
-			      height + SHADOW_MARGIN,
-			      64,
-			      64);
+		if (frame_cairo->decoration_type != DECORATION_TYPE_TILED)
+			render_shadow(cr,
+				      frame_cairo->shadow_blur,
+				      -(int)SHADOW_MARGIN/2,
+				      -(int)SHADOW_MARGIN/2,
+				      width + SHADOW_MARGIN,
+				      height + SHADOW_MARGIN,
+				      64,
+				      64);
 		break;
 	case TITLE:
 		cairo_set_rgba32(cr, &col_title);
@@ -1340,8 +1327,7 @@ draw_component_content(struct libdecor_frame_cairo *frame_cairo,
 			cairo_rel_line_to(cr, small - 1, 0);
 			cairo_rel_line_to(cr, 0, small - 1);
 			cairo_line_to(cr, x + small - 1, y + small - 1);
-		}
-		else {
+		} else {
 			cairo_rectangle(cr, x, y, SYM_DIM - 1, SYM_DIM - 1);
 		}
 		cairo_stroke(cr);
@@ -1645,6 +1631,7 @@ draw_decoration(struct libdecor_frame_cairo *frame_cairo)
 		if (is_title_bar_surfaces_showing(frame_cairo))
 			hide_title_bar_surfaces(frame_cairo);
 		break;
+	case DECORATION_TYPE_TILED:
 	case DECORATION_TYPE_ALL:
 		/* show borders */
 		ensure_border_surfaces(frame_cairo);
@@ -1658,7 +1645,7 @@ draw_decoration(struct libdecor_frame_cairo *frame_cairo)
 				&frame_cairo->plugin_cairo->visible_frame_list,
 				&frame_cairo->link);
 		break;
-	case DECORATION_TYPE_TITLE_ONLY:
+	case DECORATION_TYPE_MAXIMIZED:
 		/* hide borders */
 		if (is_border_surfaces_showing(frame_cairo))
 			hide_border_surfaces(frame_cairo);
@@ -1674,43 +1661,20 @@ draw_decoration(struct libdecor_frame_cairo *frame_cairo)
 	}
 }
 
-static void
-set_window_geometry(struct libdecor_frame_cairo *frame_cairo)
-{
-	struct libdecor_frame *frame = &frame_cairo->frame;
-	int x = 0, y = 0, width = 0, height = 0;
-
-	switch (frame_cairo->decoration_type) {
-	case DECORATION_TYPE_NONE:
-		x = 0;
-		y = 0;
-		width = libdecor_frame_get_content_width(frame);
-		height = libdecor_frame_get_content_height(frame);
-		break;
-	case DECORATION_TYPE_ALL:
-	case DECORATION_TYPE_TITLE_ONLY:
-		x = 0;
-		y = -(int)TITLE_HEIGHT;
-		width = libdecor_frame_get_content_width(frame);
-		height = libdecor_frame_get_content_height(frame) + TITLE_HEIGHT;
-		break;
-	}
-
-	libdecor_frame_set_window_geometry(frame, x, y, width, height);
-}
-
 static enum decoration_type
 window_state_to_decoration_type(enum libdecor_window_state window_state)
 {
 	if (window_state & LIBDECOR_WINDOW_STATE_FULLSCREEN)
 		return DECORATION_TYPE_NONE;
-	else if (window_state & LIBDECOR_WINDOW_STATE_MAXIMIZED ||
-		 window_state & LIBDECOR_WINDOW_STATE_TILED_LEFT ||
+	else if (window_state & LIBDECOR_WINDOW_STATE_MAXIMIZED)
+		/* title bar, no shadows */
+		return DECORATION_TYPE_MAXIMIZED;
+	else if (window_state & LIBDECOR_WINDOW_STATE_TILED_LEFT ||
 		 window_state & LIBDECOR_WINDOW_STATE_TILED_RIGHT ||
 		 window_state & LIBDECOR_WINDOW_STATE_TILED_TOP ||
 		 window_state & LIBDECOR_WINDOW_STATE_TILED_BOTTOM)
-		/* title bar, no shadows */
-		return DECORATION_TYPE_TITLE_ONLY;
+		/* title bar, invisible shadows */
+		return DECORATION_TYPE_TILED;
 	else
 		/* title bar, shadows */
 		return DECORATION_TYPE_ALL;
@@ -1754,7 +1718,6 @@ libdecor_plugin_cairo_frame_commit(struct libdecor_plugin *plugin,
 	frame_cairo->window_state = new_window_state;
 
 	draw_decoration(frame_cairo);
-	set_window_geometry(frame_cairo);
 }
 
 static void
@@ -1792,24 +1755,6 @@ libdecor_plugin_cairo_frame_property_changed(struct libdecor_plugin *plugin,
 	}
 }
 
-static void
-libdecor_plugin_cairo_frame_translate_coordinate(struct libdecor_plugin *plugin,
-						 struct libdecor_frame *frame,
-						 int content_x,
-						 int content_y,
-						 int *frame_x,
-						 int *frame_y)
-{
-	struct libdecor_frame_cairo *frame_cairo =
-		(struct libdecor_frame_cairo *) frame;
-
-	*frame_x = content_x;
-	*frame_y = content_y;
-
-	if (frame_cairo->title_bar.is_showing)
-		*frame_y += TITLE_HEIGHT;
-}
-
 static bool
 streq(const char *str1,
       const char *str2)
@@ -1877,61 +1822,38 @@ libdecor_plugin_cairo_frame_popup_ungrab(struct libdecor_plugin *plugin,
 }
 
 static bool
-libdecor_plugin_cairo_configuration_get_content_size(
-		struct libdecor_plugin *plugin,
-		struct libdecor_configuration *configuration,
-		struct libdecor_frame *frame,
-		int *content_width,
-		int *content_height)
+libdecor_plugin_cairo_frame_get_border_size(struct libdecor_plugin *plugin,
+					    struct libdecor_frame *frame,
+					    struct libdecor_configuration *configuration,
+					    int *left,
+					    int *right,
+					    int *top,
+					    int *bottom)
 {
-	int win_width, win_height;
-	if (!libdecor_configuration_get_window_size(configuration,
-						    &win_width,
-						    &win_height))
-		return false;
+	enum libdecor_window_state window_state;
 
-	enum libdecor_window_state state;
-	if (!libdecor_configuration_get_window_state(configuration, &state)) {
-		return false;
+	if (configuration) {
+		if (!libdecor_configuration_get_window_state(
+			    configuration, &window_state))
+			return false;
+	} else {
+		window_state = libdecor_frame_get_window_state(frame);
 	}
 
-	switch (window_state_to_decoration_type(state)) {
-	case DECORATION_TYPE_NONE:
-		*content_width = win_width;
-		*content_height = win_height;
-		break;
-	case DECORATION_TYPE_ALL:
-	case DECORATION_TYPE_TITLE_ONLY:
-		*content_width = win_width;
-		*content_height = win_height - TITLE_HEIGHT;
-		break;
-	}
+	if (left)
+		*left = 0;
+	if (right)
+		*right = 0;
+	if (bottom)
+		*bottom = 0;
+	if (top) {
+		enum decoration_type type = window_state_to_decoration_type(window_state);
 
-	return true;
-}
-
-static bool
-libdecor_plugin_cairo_frame_get_window_size_for(
-		struct libdecor_plugin *plugin,
-		struct libdecor_frame *frame,
-		struct libdecor_state *state,
-		int *window_width,
-		int *window_height)
-{
-	enum libdecor_window_state window_state =
-		libdecor_state_get_window_state(state);
-
-	switch (window_state_to_decoration_type(window_state)) {
-	case DECORATION_TYPE_NONE:
-		*window_width = libdecor_state_get_content_width(state);
-		*window_height = libdecor_state_get_content_height(state);
-		break;
-	case DECORATION_TYPE_ALL:
-	case DECORATION_TYPE_TITLE_ONLY:
-		*window_width = libdecor_state_get_content_width(state);
-		*window_height =
-			libdecor_state_get_content_height(state) + TITLE_HEIGHT;
-		break;
+		if (((struct libdecor_frame_cairo *)frame)->title_bar.is_showing &&
+		    (type != DECORATION_TYPE_NONE))
+			*top = TITLE_HEIGHT;
+		else
+			*top = 0;
 	}
 
 	return true;
@@ -1946,15 +1868,9 @@ static struct libdecor_plugin_interface cairo_plugin_iface = {
 	.frame_free = libdecor_plugin_cairo_frame_free,
 	.frame_commit = libdecor_plugin_cairo_frame_commit,
 	.frame_property_changed = libdecor_plugin_cairo_frame_property_changed,
-	.frame_translate_coordinate =
-		libdecor_plugin_cairo_frame_translate_coordinate,
 	.frame_popup_grab = libdecor_plugin_cairo_frame_popup_grab,
 	.frame_popup_ungrab = libdecor_plugin_cairo_frame_popup_ungrab,
-
-	.configuration_get_content_size =
-			libdecor_plugin_cairo_configuration_get_content_size,
-	.frame_get_window_size_for =
-			libdecor_plugin_cairo_frame_get_window_size_for,
+	.frame_get_border_size = libdecor_plugin_cairo_frame_get_border_size,
 };
 
 static void
@@ -2447,6 +2363,7 @@ pointer_button(void *data,
 		}
 		else if (state == WL_POINTER_BUTTON_STATE_RELEASED &&
 			 frame_cairo->grab) {
+			libdecor_frame_ref(&frame_cairo->frame);
 			if (frame_cairo->grab == frame_cairo->focus) {
 				switch (frame_cairo->active->type) {
 				case BUTTON_MIN:
@@ -2467,6 +2384,7 @@ pointer_button(void *data,
 			}
 			frame_cairo->grab = NULL;
 			sync_active_component(frame_cairo, seat);
+			libdecor_frame_unref(&frame_cairo->frame);
 		}
 	}
 	else if (button == BTN_RIGHT &&
@@ -2696,8 +2614,8 @@ output_removed(struct libdecor_plugin_cairo *plugin_cairo,
 		remove_surface_outputs(&frame_cairo->title_bar.close, output);
 	}
 	wl_list_for_each(seat, &plugin_cairo->seat_list, link) {
-		struct cursor_output *cursor_output;
-		wl_list_for_each(cursor_output, &seat->cursor_outputs, link) {
+		struct cursor_output *cursor_output, *tmp;
+		wl_list_for_each_safe(cursor_output, tmp, &seat->cursor_outputs, link) {
 			if (cursor_output->output == output) {
 				wl_list_remove(&cursor_output->link);
 				free(cursor_output);
